
/*
 *                            COPYRIGHT
 *
 *  XAnalyser, frequence spectrum analyser for X Windows
 *  Copyright (C) 1998 Arvin Schnell
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Contact addresses:
 *  aschnell@physik.uni-bremen.de or arvinschnell@yahoo.com
 *  Arvin Schnell, Am Heidberg 8, 28865 Lilienthal, Germany
 *
 */


#include <stdlib.h>
#include <iostream.h>
#include <iomanip.h>
#include <builtin.h>

#include <X11/Xlib.h>
#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/Frame.h>
#include <Xm/DrawingA.h>
#include <Xm/Protocols.h>

#include "Trans.hh"
#include "DSP.hh"
#include "fft.hh"
#include "main.hh"
#include "display.hh"
#include "control.hh"
#include "calc.hh"
#include "misc.hh"


int source;             // SOUNDCARD or DISKFILE

int channels;           // MONO, LEFT, RIGHT or STEREO

int num_channels;       // number of channels (1 or 2)

int sample_speed;       // guest what
int sample_length;      // samples per cycle (power of 2)


int num_fft;            // number of frequency-channels


short *snd_buffer;      // buffer for sound data

float *value_u[2];      // u-values of current cycle (before rfft) [sample_length]

float *value_db[2];     // dB-values of current cycle (after rfft) [num_fft]

short *value_y[2];      // y of current cycle [num_fft]
short *value_y_old[2];  // y of last cycle [num_fft]


float *value_e;         // values for envelope [sample_length]


double upper_db, lower_db;

int frame_count;
double frames_per_second;

int searchon;

float alpha;
float decay;


int marker_x[2];
int peak_x;
float peak_db;


DSP dsp;                // dsp interface
FILE *fd;



int
calc_init()
{
  // init sample data
  
  sample_speed = xanalyser.sample_speed;
  sample_length = xanalyser.sample_length;
  
  upper_db = xanalyser.upper_db;
  lower_db = xanalyser.lower_db;
  
  num_fft = sample_length / 2 + 1;                  // (yes +1)
  
  frames_per_second = (double)( sample_speed ) / sample_length;
  
  
#ifdef DEBUG
  cerr << "sample speed: " << sample_speed << " Hz\n";
  cerr << "sample length: " << sample_length << '\n';
  cerr << "frame per second: " << frames_per_second << " Hz\n";
  cerr << "num fft: " << num_fft << '\n';
  cerr << "resolution: " << sample_speed / num_fft / 2.0 << " Hz\n";
#endif
  
  
  // check sample_length
  
  int ok = 0;
  for( int i = 4; i < 16; i++ )
    if( sample_length == ( 1L << i ) )
      ok = 1;
  
  if( !ok ) {
    cerr << program_name << ": illegal sample length (must be a power of 2)\n";
    exit( 1 );
  }
  
  
  // allocate memory
  
  for( int c = 0; c < 2; c++ ) {
    value_db[c] = new float[num_fft];
    value_y[c] = new short[num_fft];
    value_y_old[c] = new short[num_fft];
    value_u[c] = new float[sample_length];
  }
  
  value_e = new float[sample_length];
  snd_buffer = new short[2*sample_length];
  
  
  // trans - other part in display_realize()
  
  K.X.setA( 0.0, sample_speed / 2 );
  K.Y.setA( lower_db, upper_db );
  
  
  // clear stuff
  
  for( int c = 0; c < 2; c++ )
    for( int i = 0; i < num_fft; i++ )
      value_db[c][i] = lower_db;
  
  
  return 1;
}



void
envelope()
{
  // normalised to 1
  
  float sum = 0.0;
  for( int i = 0; i < sample_length; i++ ) {
    float x = (float)( i - sample_length/2 ) / sample_length;
    value_e[i] = exp( - alpha * sqr( x ) );
    sum += value_e[i];
  }
  sum /= sample_length;
  
  for( int i = 0; i < sample_length; i++ )
    value_e[i] /= sum;
}



void
snd2u()
{
  const float norm = (float) MAXSHORT;
  
  switch( channels ) {
    
  case MONO:
    for( int i = 0; i < sample_length; i++ )
      value_u[0][i] = value_e[i] *
	(double)( snd_buffer[2*i+0] + snd_buffer[2*i+1] ) / norm;
    break;
    
  case LEFT:
    for( int i = 0; i < sample_length; i++ )
      value_u[0][i] = value_e[i] *
	(double)( snd_buffer[2*i+0] ) / norm;
    break;
    
  case RIGHT:
    for( int i = 0; i < sample_length; i++ )
      value_u[0][i] = value_e[i] *
	(double)( snd_buffer[2*i+1] ) / norm;
    break;
    
  case STEREO:
    for( int i = 0; i < sample_length; i++ ) {
      value_u[0][i] = value_e[i] *
	(double)( snd_buffer[2*i+0] ) / norm;
      value_u[1][i] = value_e[i] *
	(double)( snd_buffer[2*i+1] ) / norm;
      }
    break;
    
  }
}



void
calc_search_peak()
{
  peak_x = 0;
  peak_db = lower_db;
  
  for( int c = 0; c < num_channels; c++ )
    for( int i = num_fft - 1; i >= 0; i-- ) {
      
      if( i >= marker_x[0] && i <= marker_x[1] )
	if( peak_db <= value_db[c][i] ) {
	  peak_x = i;
	  peak_db = value_db[c][i];
	}
      
    }
}



void
analyse()
{
  snd2u();
  
  for( int c = 0; c < num_channels; c++ )
    rfft( value_u[c], sample_length, +1 );
  
  for( int c = 0; c < num_channels; c++ )
    for( int i = 0; i < num_fft; i++ ) {
      
      float db, db_old, p;
      
      value_y_old[c][i] = value_y[c][i];
      db_old = value_db[c][i];
      
      if( i != 0 && i != num_fft - 1 )
	p = sqr( value_u[c][2*i+0] ) + sqr( value_u[c][2*i+1] );
      else
	p = sqr( value_u[c][ i == 0 ? 0 : 1 ] );
      
      db = 10.0 * log10( p ) - 27.1;  // gauge ( 990.5273 Hz, 0 dB )
      
      if( frame_count != 0 )
	db = max( db, db_old - decay / frames_per_second );
      
      if( db > upper_db ) db = upper_db;
      if( db < lower_db ) db = lower_db;
      
      value_db[c][i] = db;
      K.Y.AtoB( value_db[c][i], value_y[c][i] );
      
    }
}



int
calc_parms()
{
  if( channels == STEREO )
    num_channels = 2;
  else
    num_channels = 1;
  
  envelope();
  
  display_remanage();
  display_marker();
  
  return 1;
}



int
calc_open()
{
  calc_parms();
  
  switch( source ) {
    
  case SOUNDCARD:
    if( !dsp.open( DSP_RECORD, AFMT_S16_LE, 1, sample_speed, 4*sample_length ) )
      return 0;
#ifdef DEBUG
    dsp.status();
#endif
    break;
    
  case DISKFILE:
    if( !dsp.open( DSP_PLAY, AFMT_S16_LE, 1, sample_speed, 4*sample_length ) )
      return 0;
#ifdef DEBUG
    dsp.status();
#endif
    if( ( fd = fopen( xanalyser.filename, "rb" ) ) == NULL ) {
      cerr << program_name << ": " << strerror( errno ) << '\n';
      dsp.close();
      return 0;
    }
    
    break;
    
  }
  
  frame_count = 0;
  
  return 1;
}



int
calc_shot()
{
  switch( source ) {
    
  case SOUNDCARD:
    dsp.read( snd_buffer, 4*sample_length );
    analyse();
    display_update();
    break;
    
  case DISKFILE:
    
    // FIXME - EOF
    
    int i = sample_length;
    
    int j;            // wieviele samples konnten geladen werden ? 
    j = fread( snd_buffer, 4, i, fd );
    
    if( j == 0 ) return 0;     // eof
    
    if( j != sample_length ) {
      return 0;
    }
    
    dsp.swapbuffer( snd_buffer, 4*sample_length );
    
    analyse();
    dsp.write( snd_buffer, 4*sample_length );
    display_update();
    
    break;
  }
  
  frame_count++;
  
  return 1;
}



int
calc_close()
{
  dsp.close();
  
  switch( source ) {
    
  case DISKFILE:
    fclose( fd );
    break;
    
  }
  
  return 1;
}

